# Copyright MIT MRC 2013

#!/usr/bin/env python
""" comfort_fuzz.py: Read in data and actually fuzz things """
__author__  = "Stelios Sidiroglou-Douskos <stelios@csail.mit.edu>"
__author__  = "Neil Fitzgerald <neilfitz@mit.edu>"
__licence__ = "MIT"

# System
import ac2_logging
import logging
import sys
from   optparse       import OptionParser
import pickle
#import glob
#import re
import os
import math
import itertools

import hachoir_daikon
from hachoir_daikon import get_file_parser, HachoirFileError, visit_field

path = lambda *paths: os.path.abspath(os.path.join(os.path.dirname(__file__), *paths)) + "/"
sys.path.append(path("/home/mrc/comfortfuzz/comfortDB/comfortDB"))
sys.path.append(path("/home/mrc/comfortfuzz/comfortDB/"))
sys.path.append(path("/home/mrc/comfortfuzz/"))
os.environ['DJANGO_SETTINGS_MODULE'] = "comfortDB.settings.prod"

from apps.comfortFuzzer.models           import Application, File, ValgrindVersion, ComfortFuzzInstance
from apps.comfortFuzzer.tasks           import FuzzTask
#from libs.url_download                  import download

from breaker.arithmetic_constraints import *
from breaker.arithmetic_utils import *


## Globals
log = None
skip_web_shivs = False

'''
    0. Read learning data
    1. Read file
    2. DFS fields
    3. For f in fields:
        - Integers:
            - Check for negative value
            - Check for upper bound
        - String:
            - Upper bound string?
            - Reencode?
    4. Create new file with all rectified fields
'''

def init_logger():
    global log
    log = ac2_logging.logger("comfort_fuzz", loglevel=logging.DEBUG)

def read_train_file(f):
    log.debug("Reading training data from %s" %f)
    return pickle.load( open(f, "rb"))

def load_input_file(f):
    log.debug("Opening %s for inspection" %f)
    try:
        file_parser = get_file_parser(f)
    except HachoirFileError:
        log.exception("Could not open/parse %s" %f)
        sys.exit(1)

    try:
        visit_field(f, file_parser, opts.use_array_names)
    except :
        log.exception("Error while parsing %s" %f)
        sys.exit(1)

    hachoir_daikon.PP.add_current_file(f)
    return file_parser

def get_field_intersection(a,b):
    return [k for k in a if k in b]

def fields_missing_from_training(a,b):
    '''
    Given two variable sets, find if there are any variables in a that are missing from b (training)
    Return True if fields are missing, False otherwise.
    '''
    return any([k for k in a if k not in b])

def get_fuzz_work(constraint_type, record, num_low_p_fuzz=5):
    l = []

    # Ask for high probability shiv results
    shivs = []
    if constraint_type.__class__.__name__ == 'StringConstraint':
        shivs = constraint_type.shiv(record.value, skip_web_shivs)
    else:
        shivs = constraint_type.shiv(record.value)
    for shiv in shivs:
        l.append( ({record.path:shiv[0]}, shiv[1]) )

    # Ask for low propability fuzz results
    if constraint_type.__class__.__name__ == 'StringConstraint':
        randoms = constraint_type.random_fuzz(record.value, num_low_p_fuzz)
        for rnd_fuzz in randoms:
            l.append( ({record.path:rnd_fuzz[0]}, rnd_fuzz[1]) )
    else:
        for i in range(num_low_p_fuzz):
            fuzz_val, score = constraint_type.random_fuzz(record.value)
            l.append( ({record.path:fuzz_val}, score) )
    ## TODO: Sort the list
    return l

def get_fuzz_work_multi(constraint_types, records, num_low_p_fuzz=5):
    assert(len(constraint_types) == len(records)) 
    low_p_fuzz_per = int(math.ceil(math.pow(num_low_p_fuzz, 1.0/len(constraint_types))))

    work_lists = []
    for i in range(len(constraint_types)):
        work_lists.append(get_fuzz_work(constraint_types[i], records[i], low_p_fuzz_per))
    master_list = list(itertools.product(*work_lists))
    l = []
    for item in master_list:
        fuzz = dict()
        total_score = 0
        for i in range(len(constraint_types)):
            fuzz[records[i].path] = item[i][0][records[i].path]
            total_score += item[i][1]
        l.append((fuzz, total_score))
    
    return l

def get_fuzz_list():
    fuzz_list = []
    
    iz3_data = pickle.load(open("iz3.p", "r"))
    variables_tried = list()
    for _, constraint_set in iz3_data.items():
        for constraint in constraint_set:
            variables = eval(constraint).get_bv_variables()
            variables = list(set(variables)) # To remove redundancy
            names = [variable.name for variable in variables]
            if names in variables_tried:
                continue
            variables_tried.append(names)
            print "Joint fuzz of the following %d variables:" % len(variables)

            constraint_types = []
            records = []

            for variable in variables:
                k = variable.name
                v = hachoir_daikon.PP.variables[k]
                print "%-30s %-15s %-30s --> %s" %(v.name, v.get_type_from_recs(), v.get_values()[:3], train.variables[k].constraint_type)
                if train.variables[k].constraint_type is not None:
                    assert(len(v.records.keys()) == 1) ## We should only at one file at a time (keys in records are the filenames)
                    for filename, record_list in v.records.items():
                        if len(record_list) > 1: log.debug("%s has %d records" %(v.name, len(record_list)))
                        for record in record_list:
                            constraint_types.append(train.variables[k].constraint_type)
                            records.append(record)
                else:
                    log.debug("%s has no constraint type associated with it" %v.name)

            fuzz_list.extend(get_fuzz_work_multi(constraint_types, records))
            print 80 * '-'

    for k,v in hachoir_daikon.PP.variables.items():
        if train.variables.has_key(k):
            print "%-30s %-15s %-30s --> %s" %(v.name, v.get_type_from_recs(), v.get_values()[:3], train.variables[k].constraint_type)
            if train.variables[k].constraint_type is not None:
                assert(len(v.records.keys()) == 1) ## We should only at one file at a time (keys in records are the filenames)
                for filename, record_list in v.records.items():
                    if len(record_list) > 1: log.debug("%s has %d records" %(v.name, len(record_list)))
                    for record in record_list:
                        fuzz_list.extend(get_fuzz_work(train.variables[k].constraint_type, record))
            else:
                log.debug("%s has no constraint type associated with it" %v.name)
        else:
            log.warning("[-] No match found for %s in training set" %v.name)

    ## Hack to clear duplicates and order the list
    seen = set()
    uniq_fuzz_list = []
    for d,s in fuzz_list:
        t = tuple(d.items())
        if t not in seen:
            seen.add(t)
            uniq_fuzz_list.append((d,s))
    return sorted(uniq_fuzz_list, key=lambda k: k[1], reverse=True)

def parse_opts():
    parser = OptionParser(usage = "usage: %prog [options]")

    parser.add_option("-l", "--train-input", dest = "train_file",
                      action  = "store",
                      default = "train.p",
                      help    = "Training pickle file")

    parser.add_option("-A", "--array_name", dest = "use_array_names",
                    action  = "store_true",
                    default = False,
                    help    = "Flag to use array descriptions as indexes (instead of \"..\")")

    parser.add_option(
        "--app-name", dest="app_name",
        action      = "store",
        #default     = "pngTestReader",
        default     = "pngCleanOverflowMultiMalloc",
        help        = "Application name"
    )
    parser.add_option(
        "--app-version", dest="app_version",
        action      = "store",
        default     = "0.1",
        help        = "Application name"
    )
    parser.add_option(
        "--app-path", dest="app_path",
        action      = "store",
        default     = None,
        help        = "Application path"
    )
    parser.add_option(
        "--app-params", dest="app_params",
        action      = "store",
        default     = None,
        help        = "Application params"
    )
    parser.add_option(
        "-f", "--seed-file", dest="seed_file",
        action      = "store",
        default     = "2YL9OobZmyLA7dSJV_qzFA.png",
        help        = "Seed file to start testing on"
    )
    parser.add_option(
        "--seed-file-url", dest="url_path",
        action      = "store",
        default     = None,
        help        = "Url path to download seed file from"
    )
    parser.add_option(
        "--seed-file-format", dest="file_format",
        action      = "store",
        default     = None,
        help        = "Seed file format"
    )

    parser.add_option(
        "-L", "--log-level", dest="loglevel",
        action      = "store",
        default     = "info",
        help        = "Set log level output. (e.g., --log-level DEBUG)"
    )

    parser.add_option(
        "--interactive", dest="query",
        action      = "store_true",
        default     = False,
        help        = "Interactively query user whether to fuzz a particular program point"
    )

    parser.add_option(
        "--raw-bytes", dest="use_raw_bytes",
        action      = "store_true",
        default     = False,
        help        = "Use raw bytes instead of symbolic hachoir paths"
    )
    parser.add_option(
        "--gold-standard", dest="use_gold_std",
        action      = "store_true",
        default     = False,
        help        = "Use gold standard to detect errors"
    )
    parser.add_option(
        "-J", "--ignore-jumps", dest="ignore_jumps",
        action      = "store_true",
        default     = False,
        help        = "Ignore jumps when running Fluore for tracing calculations"
    )

    parser.add_option(
        "--skip-file-removal", dest="remove_tmp_files",
        action      = "store_false",
        default     = True,
        help        = "Don't remove tmp files. Useful for debugging."
    )

    parser.add_option(
        "--stop-on-unmmap", dest="stop_on_unmmap",
        action      = "store",
        default     = None,
        help        = "Exit program when we unmmmap taint file N times"
    )

    parser.add_option(
        "--timeout", dest="timeout",
        action      = "store",
        default     = "10m",
        help        = "Time to wait until we timeout "
    )

    parser.add_option(
        "--local-worker", dest="run_locally",
        action      = "store_true",
        default     = False,
        help        = "Flag to use send tasks to remote workers"
    )

    parser.add_option(
        "--extra-fluore-args", dest="extra_fluore_args",
        action      = "store",
        default     = None,
        help        = "extra arguments to fluore (white spaced separated)"
    )

    parser.add_option(
        "--dies-with-signal-timeout", dest="dies_with_signal_timeout",
        action      = "store",
        default     = "1m",
        help        = "timeout for testing mutated files"
    )

    parser.add_option(
        "--skip-web-shivs", dest="skip_web_shivs",
        action      = "store_true",
        default     = False,
        help        = "Skip JavaScript/URL-specific string shivs"
    )

    return parser

##XXX: Copied from work generator (TODO: import from work_generator module)
def get_app_by_name(app_name):
    """
    @param      app_name: Name of application
    @param      app_ver: Application version
    @return:    Application object
    """
    try:
        return Application.objects.all().filter(name=app_name)[0]
    except IndexError:
        return None

def create_cfi(cfi_ver, fluore_ver, constr_breaker, valgrind_ver):
    # Get Valgrind version
    if not ValgrindVersion.objects.all().filter(version=VALGRIND_VERSION).exists():
        log.debug( "Creating new valgrind version")
        valgrind = ValgrindVersion(version=VALGRIND_VERSION)
        valgrind.save()
    else:
        log.debug( "Getting valgrind version from DB")
        valgrind = ValgrindVersion.objects.all().filter(version=VALGRIND_VERSION)[0]
    log.debug( "Valgrind version:%s" %valgrind.version)

    # Get CFI
    log.debug( "Creating CFI")
    cfi = ComfortFuzzInstance(
        version=COMF_FUZZ_INST_VER, fluore_version=FLUORE_VER, \
        constraint_breaker_version=CONST_BRKR_VER, \
        valgrindVersion=valgrind
    )
    cfi.save()
    return cfi


if __name__ == "__main__":
    parser        = parse_opts()
    (opts, args)  = parser.parse_args(None)

    ## Initialize logger
    init_logger()

    ## Load training data
    train = read_train_file(opts.train_file)

    ## Load suspicious file
    fp = load_input_file(unicode(opts.seed_file))

    ## Get ordered fuzz work list
    skip_web_shivs = opts.skip_web_shivs
    uniq_fuzz_list = get_fuzz_list()

    log.info("Generated fuzz list of %d items" %len(uniq_fuzz_list))
    for item in uniq_fuzz_list:
        print item

    VALGRIND_VERSION   = "valgrind-3.8.1"
    COMF_FUZZ_INST_VER = "0.0.1"
    FLUORE_VER         = "Fluore-0.0.1"
    CONST_BRKR_VER     = "constraint_breaker-0.0.1"

    ## Check flag to determine if we want to use cloud workers or run locally
    if opts.run_locally:
        ## I had to add the name "apps" because of the way the task in celery pick up their logger
        log           = ac2_logging.logger("apps", loglevel = getattr(logging, opts.loglevel.upper(), None))
        log.info("Using local workers.")

        from celery import current_app
        current_app.conf.CELERY_ALWAYS_EAGER                = True
        current_app.conf.CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
    else:
        log           = ac2_logging.logger("comfort_fuzz", loglevel = getattr(logging, opts.loglevel.upper(), None))
        log.info("Using remote workers. Make sure that they are ready to accept jobs.")
    #
    # Enable the option to operate without the DB.
    #
    if opts.app_path is not None and opts.app_params is not None and opts.url_path is not None and opts.file_format is not None:
        ## SKip the database
        ##XXX -app-name, --seed_file should be off here.
        log.info("Using user-supplied application (i.e., not the one in the database)")
        app_path    = opts.app_path
        app_params  = opts.app_params
        url_path    = opts.url_path
        file_format = opts.file_format
        log.info("path:%s, params:%s, url_path:%s, file_format:%s", app_path, app_params, url_path, file_format)
    else:
        ## Use the database

        ## Application
        app = get_app_by_name(opts.app_name)
        if app is None:
            log.error("Could not find application:%s:%s" %(opts.app_name, opts.app_version))
            sys.exit(1)
        app_path, app_params    = app.path, app.params

        ## ComfortFuzzInstance
        if not ComfortFuzzInstance.objects.all().filter(version=COMF_FUZZ_INST_VER).exists():
            cfi = create_cfi(COMF_FUZZ_INST_VER, FLUORE_VER, CONST_BRKR_VER, VALGRIND_VERSION)
        else:
            log.debug( "Using CFI from DB")
            cfi = ComfortFuzzInstance.objects.all().filter(version=COMF_FUZZ_INST_VER)[0]

        ## Seed File
        file_name = opts.seed_file
        if not File.objects.all().filter(name=file_name).exists():
            log.error( "Could not access file at %s" %file_name)
            sys.exit(1)
        seed_file   = File.objects.all().get(name=file_name)
        url_path    = seed_file.url_path
        file_format = seed_file.format


    results=[]
    ## Generate work
    for f in uniq_fuzz_list:
        assignments = f[0]
        log.debug("Going to fuzz %s with %s" %(app_path, assignments))
        fz  = FuzzTask()
        res = fz.delay(
            url_path,
            app_path,
            app_params,
            file_format,
            assignments,
            opts.dies_with_signal_timeout,
        )
        results.append((assignments,res))

    log.info("Waiting for %d results to return" %(len(results)))
    for assignments, res in results:
        try:
            res.wait()
        except:
            log.exception("Error getting result for %s", assignments)

    bugs=[]
    for assignments, res in results:
        if not res.successful():
            log.error("Error fuzzing %s" %assignments)
            continue
        found_fault, assignments = res.result
        if found_fault:
            log.info("Found fault with assignments: %s" %assignments)
            bugs.append(assignments)

    ##TODO: Result binning.
    log.info("Found %d bugs" %len(bugs))
    print bugs
