# Copyright MIT MRC 2013

__author__  = "Nathan Rittenhouse <nathan_@mit.edu>"
__author__  = "Neil Fitzgerald <neilfitz@mit.edu>"
__licence__ = "MIT"

from common import *
from fuzz_common import *
from urlparse import urlsplit, urlunsplit
import copy

evil_netloc = "31.33.07.90:8080"
evil_path = "meow.jpg.exe"

def until_not_avoided(tries, function, avoid_strategy, *packaging):
	for i in range(0, tries):
		new_fuzz_value = function(*packaging)
		
		if new_fuzz_value in avoid_strategy.avoid_values:
			continue
		
		return new_fuzz_value
	
	return None

class StringModifier(object):
	def __init__(self, start_string, info):
		self.string = copy.copy(start_string)
		self.old_string = start_string
		self.old_length = len(self.string)
		
		if "encoding" in info:
			self.encoding = info["encoding"]
		else:
			self.encoding = "ascii"
			
		self.allow_expand = False
		
		self.raw_mode = False
		self.mode = None
		

	# Are we allowed to expand the string or not?
	def allow_expansion(self):
		self.allow_expand = True
		
	def disallow_expansion(self):
		self.allow_expand = False
		
	# Change the current encoding assumed for inserted values
	def set_encoding(self, encoding):
		self.encoding = encoding

	def encode_value(self, value):
		if not self.raw_mode:
			return value.encode(self.encoding)
		return value

	def raw(self):
		self.raw_mode = True
		return self

	def unset_raw(self):
		self.raw_mode = False

	def enforce_expansion(self):
		# Assume self.old_length is the value for the old string length
		if self.allow_expand:
			self.old_length = len(self.string)
			return

		if len(self.string) > self.old_length:
			self.string = self.string[:self.old_length]

		self.old_length = len(self.string)


	def end_run(self):
		self.current_pos = None

	def start_run(self, index):
		self.current_pos = index

	# "wedge" a value into the string rather than overwrite all char
	def insert_after(self, index, value):
		self.end_run()
		encoded = self.encode_value(value)
		
		self.string = self.string[:index+1] + encoded + self.string[index+2:]
		
		self.start_run(index + len(encoded)+2)

		self.enforce_expansion()
		self.unset_raw()

	def null(self):
		if self.encoding == "utf-8" or self.encoding == "ascii":
			return "\x00"
		elif self.encoding == "utf-16":
			return "\x00\x00"
		elif self.encoding == "utf-32":
			return "\x00\x00\x00\x00"
		
	def insert(self, index, value):
		self.end_run()
		self.mode = "insert"
		encoded = self.encode_value(value)
		
		self.string = self.string[:index] + encoded + self.string[index+1:]

		self.start_run(index + len(encoded)+1)

		self.enforce_expansion()
		self.unset_raw()

	def overwrite(self, index, value):
		self.end_run()
		self.mode = "overwrite"
		encoded = self.encode_value(value)

		self.string = self.string[:index] + encoded + self.string[index+len(encoded):]

		self.start_run(index + len(encoded))

		self.enforce_expansion()
		self.unset_raw()

	def overwrite_after(self, index, value):
		self.end_run()
		self.mode = "overwrite"
		encoded = self.encode_value(value)

		self.string = self.string[:index+1] + encoded + self.string[index+1+len(encoded):]
		self.start_run(index + len(encoded))

		self.enforce_expansion()
		self.unset_raw()

	def run_append(self, value):
		encode = self.encode_value(value)

		if self.mode == "insert":
			self.insert(self.current_pos, value)
		elif self.mode == "overwrite":
			self.overwrite(self.current_pos, value)
		else:
			raise Exception("No run has been entered yet.")

	def append(self, value):
		encoded = self.encode_value(value)

		self.string = self.string + encoded

		self.start_run(len(self.string)-1)

		self.enforce_expansion()
		self.unset_raw()

	def find(self, value):
		encoded = self.encode_value(value)

		return self.string.find(encoded)
	
	def reset(self):
		self.string = copy.copy(self.old_string)
		self.old_length = len(self.string)

	def __len__(self):
		return len(self.string)

def fill_repeat_string(repeat_string, base, info):
	# print "node.length:", node.length
	# print "node.value:", node.value

	counter = 0
	fuzz_string = ""

	m = StringModifier(repeat_string, info)

	# repeat_string = repeat_string.encode(info['encoding'])

	length_repeat_string = len(m)
	# format_length = node.length - (node.length % length_repeat_string)
	format_length = len(base) - (len(base) % length_repeat_string)
	
	while counter < format_length:
		fuzz_string += m.string
		counter += length_repeat_string

	# return fuzz_string + node.value[len(fuzz_string):len(node.value)]
	return fuzz_string + base[len(fuzz_string):]

def substitute_string(prefix_string, base, append_null=False, info={}):
	# print "node.length:", node.length
	# print "node.value:", node.value
	m = StringModifier(prefix_string, info)

	if append_null:
		m.raw().append(m.null())
		# prefix_string = prefix_string + "\x00".encode(info["encoding"])

	length_prefix_string = len(m)

	# if node.length < length_prefix_string:
	if len(base) < length_prefix_string:
		return base

	# return prefix_string + node.value[length_prefix_string:]
	return prefix_string + base[length_prefix_string:]

def numfuzz_random(start_value, seed_values, avoid_strategy, info):
	def func(start_value, seed_values, info):
		return BitArray(uint=randint(0, 2**info['bit_width']), length=info['bit_width']).uint
		
	return until_not_avoided(5, func, avoid_strategy, start_value, seed_values, info)

def numfuzz_signed(start_value, seed_values, avoid_strategy, info):
	value = BitArray(uint=0, length=info['bit_width'])
	value.overwrite("0b1", 0)
	return value.uint

def numfuzz_large_unsigned(start_value, seed_values, avoid_strategy, info):
	value = BitArray(uint=0, length=info['bit_width'])
	value.set("0b1", range(1, info['bit_width']))
	return value.uint

def numfuzz_negative_one(start_value, seed_values, avoid_strategy, info):
	value = BitArray(uint=0, length=info['bit_width'])
	value.set("0b1", range(0, info['bit_width']))
	return value.uint

def numfuzz_zero(start_value, seed_values, avoid_strategy, info):
	value = BitArray(uint=0, length=info['bit_width'])
	return value.uint

def numfuzz_off_by_one(start_value, seed_values, avoid_strategy, info):
	for value in seed_values:
		value_plus = value + 1
		value_minus = value - 1

		# if not (value_plus in path_values):
		if not (value_plus in avoid_strategy.avoid_values):
			if NumberManipulation.value_fits(value_plus, info['bit_width']):
				return value_plus

		if not (value_minus in avoid_strategy.avoid_values):
			if NumberManipulation.value_fits(value_minus, info['bit_width']):
				return value_minus
	return None

def numfuzz_random_alternate(start_value, seed_values, avoid_strategy, info):
	tried_values = 0
	
	while tried_values < len(seed_values):
		fuzz_value = choice(seed_values)
		if fuzz_value in avoid_strategy.avoid_values:
			tried_values += 1
		else:
			return fuzz_value
		
	return None

def numfuzz_twos_complement(start_value, seed_values, avoid_strategy, info):
	twos_complement = NumberManipulation.twos_complement(start_value, info["bit_width"])
	
	if twos_complement not in avoid_strategy.avoid_values:
		return twos_complement
	
	# Try twos-complementing the seed values
	for value in seed_values:
		complement = NumberManipulation.twos_complement(value, info["bit_width"])
		if complement not in avoid_strategy.avoid_values:
			return complement
		
	# Try twos-complementing the avoided values
	# print "avoid_values:", str(avoid_strategy.avoid_values)
	for value in avoid_strategy.avoid_values:
		complement = NumberManipulation.twos_complement(value, info["bit_width"])
		if complement not in avoid_strategy.avoid_values:
			return complement
		
	return None

def numfuzz_violate_range(start_value, seed_values, avoid_strategy, info):
	largest_value = NumberManipulation.largest_integer(info["bit_width"])

	range_choice = randint(0,1)

	force_fuzz_upper = False
	force_fuzz_lower = False

	if info["lower_bound"] <= 1:
		# Only fuzz the upper bound if the lower bound is non-existent
		force_fuzz_upper = True
		range_choice = 1
	if info["upper_bound"] >= largest_value-1:
		force_fuzz_lower = True
		range_choice = 0

	if force_fuzz_upper and force_fuzz_lower:
		raise Exception("[numfuzz_violate_range] got bounds that obviate all possible fuzz values!")

	while True:
		# Break the lower bound constraint
		if range_choice == 0:
			print "fuzzing lower bound"
			fuzz_value = randint(0, info["lower_bound"]-1)
			print "fuzz_value:", fuzz_value
		elif range_choice == 1:
			print "fuzzing upper bound"
			fuzz_value = randint(info["upper_bound"]+1, largest_value)

		if fuzz_value not in avoid_strategy.avoid_values:
			return fuzz_value

	return None

def strfuzz_use_other_variable(start_value, seed_values, avoid_strategy, info):
	for alternate_value in seed_values:
		# if (alternate_value not in avoid_strategy.avoid_values) and alternate_value != start_value:
		
		if alternate_value not in avoid_strategy.avoid_values:
			return alternate_value
	return None

def strfuzz_duplicate(start_value, format_string, base_number, info):
	m = StringModifier(start_value, info)
	
	if m.string.endswith(m.null()):
		raw_string = m.string[:-len(m.null())]
	else:
		raw_string = m.string
		
	if "max_length" in info:
		dup_count = info["max_length"]/len(start_value) + randint(0, 10)
	else:
		dup_count = randint(0, 10)

	return start_value*dup_count

def strfuzz_truncate(start_value, format_string, base_number, info):
	if "min_length" in info:
		new_length = randint(0, info["min_length"])
	else:
		new_length = randint(0, len(start_value))

	return start_value[:new_length]
		
def string_fuzz_number(start_value, format_string, base_number, info):
	pairs = [(base_number, 2000), ((base_number >> 32), 200), ((base_number >> 48), 20), ((base_number >> 56), 20)]
	
	for (base, upper) in pairs:
		if len(start_value) >= len(format_string % base):
			fuzz_value = base - randint(0, upper)
			return substitute_string(start_value, format_string % fuzz_value)
		
	return start_value

def strfuzz_format_string_number(start_value, seed_values, avoid_strategy, info):
	return fill_repeat_string("%n", start_value, info)

def strfuzz_format_string_string(start_value, seed_values, avoid_strategy, info):
	return fill_repeat_string("%s", start_value, info)

def strfuzz_fill_null(start_value, seed_values, avoid_strategy, info):
	return fill_repeat_string("\x00", start_value, info)

def strfuzz_strip_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	index = m.raw().find(m.null())

	# Overwrite a null with a Z
	if index >= 0:
		m.overwrite(index, "z")
	else:
		return start_value
	
def strfuzz_xss_script_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "<script>alert(0)</script>")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_xss_script(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "<script>alert(0)</script>")
	return m.string

def strfuzz_xss_noscript_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "alert(0)")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_xss_noscript(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "alert(0)")
	return m.string

def strfuzz_xss_url_jpg(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "file://s/i.jpg")
	return m.string

def strfuzz_xss_url_jpg_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "file://s/i.jpg")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_evil_url(start_value, seed_values, avoid_strategy, info, use_null=False):
	if "url_range" in info:
		begin, end = info["url_range"]
		
		url = start_value[begin:end+1].decode(info['encoding'])
		scheme, netloc, path, querystring, fragment = urlsplit(url)

		netloc = evil_netloc
		path = evil_path

		evil_url = urlunsplit((scheme, netloc, path, querystring, fragment))
		m = StringModifier(start_value, info)
		m.overwrite(begin, evil_url)
		
		if use_null:
			m.raw().run_append(m.null())
		
		return m.string
	return None

def strfuzz_evil_url_nonull(start_value, seed_values, avoid_strategy, info):
	return strfuzz_evil_url(start_value, seed_values, avoid_strategy, info, False)

def strfuzz_evil_url_null(start_value, seed_values, avoid_strategy, info):
	return strfuzz_evil_url(start_value, seed_values, avoid_strategy, info, True)
		
def strfuzz_html_url_jpg_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "<img src=\"file://s/i.jpg\"/>" + m.null())
	m.raw().run_append(m.null())
	return m.string

def strfuzz_command_inject_ping_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "ping s")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_command_inject_ping(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "ping s")
	return m.string

def strfuzz_html_url_jpg(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "<img src=\"file://s/i.jpg\"/>")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_library_inject_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "s.so")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_library_inject(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "s.so")
	return m.string

def strfuzz_xss_url_js(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "file://s/i.js")
	return m.string

def strfuzz_xss_url_js_null(start_value, seed_values, avoid_strategy, info):
	m = StringModifier(start_value, info)
	m.overwrite(0, "file://s/i.js")
	m.raw().run_append(m.null())
	return m.string

def strfuzz_largest_signed_hex(start_value, seed_values, avoid_strategy, info):
	return string_fuzz_number(start_value, "0x%x", 0xffffffffffffffff, info)

def strfuzz_largest_signed_dec(start_value, seed_values, avoid_strategy, info):
	return string_fuzz_number(start_value, "%d", 0xffffffffffffffff, info)

def strfuzz_largest_negative_unsigned_dec(start_value, seed_values, avoid_strategy, info):
	return string_fuzz_number(start_value, "-%d", 0x7ffffffffffffffff, info)

class NumberFuzzer(object):
	def __init__(self, start_number, info, seed_values=[], avoid_strategy=None, choice_strategy=None):
            self.strategies = [{"name":"signed", "function":numfuzz_signed, "non_deterministic": False, "rating":100000},
                               {"name":"large_unsigned", "function":numfuzz_large_unsigned, "non_deterministic": False, "rating":100000},
                               {"name":"negative_one", "function":numfuzz_negative_one, "non_deterministic": False, "rating":100000},
			       {"name":"twos_complement", "function":numfuzz_twos_complement, "non_deterministic": False, "rating":10000},
                               {"name":"zero", "function":numfuzz_zero, "non_deterministic": False, "rating":100000},
			       {"name":"off_by_one", "function":numfuzz_off_by_one, "non_deterministic": True, "rating": 10000},
			       {"name":"alternate", "function":numfuzz_random_alternate, "non_deterministic": True, "rating": 10000},
                               {"name":"random", "function":numfuzz_random, "non_deterministic": True, "rating":100},
			       {"name":"violate_range", "function":numfuzz_violate_range, "non_deterministic": True, "rating":100000}]

	    # Values that can be used as a basis for generating fuzz values
            self.seed_values = seed_values
	    self.start_number = start_number

	    # The maximum number of times that we can run a fuzzer before we decide to give up on it
	    # for the current selection
	    self.max_run_count = 30

	    # Avoid using the range breaker strategy by default.
	    def strategy_avoid_range_breaker(self, strategies):
	    	    while True:
	    		    chosen = choice(strategies)
	    		    if chosen['name'] != 'break_range':
	    			    return chosen

	    self.avoid_strategy = avoid_strategy
	    self.choice_strategy = choice_strategy
	    
	    # if avoid_strategy == None:
	    # 	    self.avoid_strategy = Avoider(avoider_use_once, [], [])
	    # else:
	    # 	    self.avoid_strategy = avoid_strategy

	    # if choice_strategy == None:
	    # 	    
	    # 	    self.choice_strategy = Choice(selector_avoid_range_breaker)
	    # else:
	    # 	    self.choice_strategy = choice_strategy
	    
	    self.info = info
	    self.info["initializing"] = True

            self.strategies[:] = [strategy for strategy in self.strategies if (strategy["name"] not in self.avoid_strategy.avoid_strategies) or self.avoid_strategy(strategy, None, self.info)]

	    self.info["initializing"] = False
	    

	def add_info(info):
            self.info = dict(self.info.items() + info.items())

	def set_choice_strategy(self, selector):
            return selector(self.strategies)

        def set_avoid_strategy(self, avoid_strategy):
	    self.avoid_strategy = avoid_strategy

	def update_strategies(self):
		for strategy_name in self.avoid_strategy.avoid_strategies:
			try:
				for mutator in self.string_mutators:
					if mutator["name"] == strategy_name:
						self.string_mutators.remove(mutator)
						break
			except:
				continue

	def fuzz(self):
		strategy = {"name":None}
		retval = None
		
		while True:
			self.update_strategies()

			# If we can't get a strategy, just return
			try:
				strategy = self.choice_strategy(self.strategies, self.info)
			except:
				return (None, None)

			
			
			run_count = 0
			terminated = False

			while (not terminated) and (run_count < self.max_run_count):
				
				value = strategy["function"](self.start_number, self.seed_values, self.avoid_strategy, self.info)
				retval = value, strategy

				if value not in self.avoid_strategy.avoid_values:
					terminated = True
					
				self.avoid_strategy(strategy, value, self.info)
				run_count += 1

			if not terminated:
				self.info["timed_out"] = True
				self.avoid_strategy(strategy, value, self.info)
				self.info["timed_out"] = False
			else:
				return retval
			
		return (None, None)
	    
class StringFuzzer(object):
	def __init__(self, start_string, info, seed_values=[], avoid_strategy=None, choice_strategy=None):
		self.start_string = start_string
		self.seed_values = seed_values

		self.max_run_count = 30

	        def use_once(self, strategy, value, info):
                    name = strategy["name"]
		    if name != "off_by_one":
			    self.avoid_strategies.append(name)
		    else:
			    self.avoid_values.append(value)

		if avoid_strategy == None:
		    self.avoid_strategy = Avoider(use_once, [], [])
		else:
		    self.avoid_strategy = avoid_strategy

		if choice_strategy == None:
		    self.choice_strategy = Chooser(choice)
		else:
	            self.choice_strategy = choice_strategy
		    
		self.string_mutators = [
			{"name":"largest_negative_unsigned_dec", "function":strfuzz_largest_negative_unsigned_dec, "non_deterministic": False, "rating":10000},
			{"name":"largest_signed_dec", "function":strfuzz_largest_signed_dec, "non_deterministic": False, "rating": 10000},
		        {"name":"largest_signed_hex", "function":strfuzz_largest_signed_hex, "non_deterministic": False, "rating":10000},
         		{"name":"xss_url_js_null", "function":strfuzz_xss_url_js_null, "non_deterministic": False, "rating":10000},
			{"name":"xss_url_js", "function":strfuzz_xss_url_js, "non_deterministic": False, "rating":10000},
			{"name":"library_inject", "function":strfuzz_library_inject, "non_deterministic": False, "rating":10000},
			{"name":"library_inject_null", "function":strfuzz_library_inject_null, "non_deterministic": False, "rating":10000},
			{"name":"html_url_jpg", "function":strfuzz_html_url_jpg, "non_deterministic": False, "rating":10000},
			{"name":"command_inject_ping", "function":strfuzz_command_inject_ping, "non_deterministic": False, "rating":10000},
			{"name":"command_inject_ping_null", "function":strfuzz_command_inject_ping_null, "non_deterministic": False, "rating":10000},
			{"name":"html_url_jpg_null", "function":strfuzz_html_url_jpg_null, "non_deterministic": False, "rating":10000},
			{"name":"xss_url_jpg_null", "function":strfuzz_xss_url_jpg_null, "non_deterministic": False, "rating":10000},
			{"name":"xss_url_jpg", "function":strfuzz_xss_url_jpg, "non_deterministic": False, "rating":10000},
			{"name":"xss_noscript_null", "function":strfuzz_xss_noscript_null, "non_deterministic": False, "rating":10000},
			{"name":"evil_url_nonull", "function":strfuzz_evil_url_nonull, "non_deterministic": False, "rating":100000},
			{"name":"evil_url", "function":strfuzz_evil_url_null, "non_deterministic": False, "rating":100000},
			{"name":"xss_noscript", "function":strfuzz_xss_noscript, "non_deterministic": False, "rating":10000},
			{"name":"strip_null", "function":strfuzz_strip_null, "non_deterministic": False, "rating":10000},
			{"name":"fill_null", "function":strfuzz_fill_null, "non_deterministic": False, "rating":10000},
			{"name":"format_string_string", "function":strfuzz_format_string_string, "non_deterministic": False, "rating":10000},
			{"name":"format_string_number", "function":strfuzz_format_string_number, "non_deterministic": False, "rating":10000},
			{"name":"use_other_variable", "function":strfuzz_use_other_variable, "non_deterministic": True, "rating":10000},
			{"name":"duplicate", "function":strfuzz_duplicate, "non_deterministic": True, "rating":10000},
			{"name":"truncate", "function":strfuzz_truncate, "non_deterministic": True, "rating":10000}]
		
		self.information = info
		self.information["initializing"] = True
		self.information["timed_out"] = False

		self.string_mutators[:] = [mutator for mutator in self.string_mutators if (mutator["name"] not in self.avoid_strategy.avoid_strategies) or self.avoid_strategy(mutator, None, self.information) ]

		self.information["initializing"] = False

		
		
	def info(information):
		self.information = information
		return self
	
	def update_strategies(self):
		for strategy_name in self.avoid_strategy.avoid_strategies:
			try:
				for mutator in self.string_mutators:
					if mutator["name"] == strategy_name:
						self.string_mutators.remove(mutator)
						break
			except:
				continue
	
	def fuzz(self):
		# TODO: handle different types of encoding of the format string!
		if len(self.start_string) == 0:
			return None
		
		strategy = {"name":None}
		retval = None

		while True:
			self.update_strategies()
			# If we can't get a strategy, just return
			try:
				strategy = self.choice_strategy(self.string_mutators, self.information)
			except:
				return (None, None)
			
			run_count = 0
			terminated = False

			while (not terminated) and (run_count < self.max_run_count):
				value = strategy["function"](self.start_string, self.seed_values, self.avoid_strategy, self.information)
				retval = value, strategy

				if value not in self.avoid_strategy.avoid_values:
					terminated = True
					
				self.avoid_strategy(strategy, value, self.information)
				run_count += 1

			

			if not terminated:
				self.information["timed_out"] = True
				self.avoid_strategy(strategy, value, self.information)
				self.information["timed_out"] = False
			else:
				return retval
		return (None, None)

				
	# def break_strategy_flip_signedness(self, path):
	# 	path_values = self.get_values(path)

	# 	for value in path_values:
			
	# Returns None when out of strings to fuzz
	def get_fuzzed_string(self):
		pass
